import pytest
import torch

from interp.circuit.computational_node import normalize_index
from rust_circuit import Array, Cumulant, DiscreteVar, GeneralFunction, RunDiscreteVarAllSpec, Sampler
from rust_circuit.py_utils import make_index_at


def test_gen_index_at_batch_x():
    torch.manual_seed(1324)
    x_shape = [1, 4, 2, 5, 7]
    index_shape = [1, 4, 2]
    for index_dim in [0, 1, -2, -1]:
        suffix_shape = x_shape[len(index_shape) :]
        use_index_dim = normalize_index(index_dim, len(suffix_shape))
        suffix_post_index_shape = suffix_shape[:use_index_dim] + suffix_shape[use_index_dim + 1 :]
        x_raw = torch.randn(*x_shape, dtype=torch.float64)
        index_raw = torch.randint(low=0, high=x_shape[use_index_dim + len(index_shape)], size=index_shape)

        x = Array(x_raw, name="x")
        index = Array(index_raw.long(), name="index")

        indexed = GeneralFunction.gen_index(x=x, index=index, index_dim=index_dim, batch_x=True)
        evaluated = indexed.evaluate()
        assert evaluated.shape == indexed.shape

        answer = torch.zeros(*index_shape, *suffix_post_index_shape, dtype=torch.float64)
        assert answer.shape == evaluated.shape
        for i in range(index_shape[0]):
            for j in range(index_shape[1]):
                for k in range(index_shape[2]):
                    answer[i, j, k] = x_raw[
                        (i, j, k) + make_index_at(index_raw[i, j, k], use_index_dim)
                    ]  # of course this doesn't generalise to e.g hypothesis tests : )

        torch.testing.assert_close(answer, evaluated)


def test_gen_index_at_batch_x_discrete_var():
    torch.manual_seed(1324)
    x_shape = [12, 4, 2, 5, 7]
    index_shape = [12, 4, 2]
    for index_dim in [0, 1, -2, -1]:
        suffix_shape = x_shape[len(index_shape) :]
        use_index_dim = normalize_index(index_dim, len(suffix_shape))
        suffix_post_index_shape = suffix_shape[:use_index_dim] + suffix_shape[use_index_dim + 1 :]

        x_raw = torch.randn(*x_shape, dtype=torch.float64)
        index_raw = torch.randint(low=0, high=x_shape[use_index_dim + len(index_shape)], size=index_shape)

        x = Array(x_raw, name="x")
        index = Array(index_raw, name="index")

        var_x = DiscreteVar(x, name="var_x")
        var_index = DiscreteVar(index, name="var_index", probs_and_group=var_x.probs_and_group)
        indexed = GeneralFunction.gen_index(x=var_x, index=var_index, index_dim=index_dim, batch_x=True)

        cumulant = Cumulant(indexed, name="cumulant")

        evaluated = (
            Sampler(
                RunDiscreteVarAllSpec(
                    (var_x.probs_and_group,),
                )
            )
            .estimate(cumulant)
            .evaluate()
        )
        assert evaluated.shape == indexed.shape

        answer = torch.zeros(*index_shape, *suffix_post_index_shape, dtype=torch.float64)
        assert answer[0].shape == evaluated.shape
        for i in range(index_shape[0]):
            for j in range(index_shape[1]):
                for k in range(index_shape[2]):
                    answer[i, j, k] = x_raw[
                        (i, j, k) + make_index_at(index_raw[i, j, k], use_index_dim)
                    ]  # of course this doesn't generalise to e.g hypothesis tests : )

        mean_answer = answer.mean(0).cpu()
        torch.testing.assert_close(mean_answer, evaluated.cpu())


def test_gen_index_at_no_batch_x():
    torch.manual_seed(1324)
    x_shape = [5, 7]
    index_shape = [1, 4, 2]
    for index_dim in [0, 1, -2, -1]:
        use_index_dim = normalize_index(index_dim, len(x_shape))
        post_index_shape = x_shape[:use_index_dim] + x_shape[use_index_dim + 1 :]
        x_raw = torch.randn(*x_shape, dtype=torch.float64)
        index_raw = torch.randint(low=0, high=x_shape[use_index_dim], size=index_shape)

        x = Array(x_raw, name="x")
        index = Array(index_raw, name="index")

        indexed = GeneralFunction.gen_index(x=x, index=index, index_dim=index_dim, batch_x=False)
        evaluated = indexed.evaluate()
        assert evaluated.shape == indexed.shape

        answer = torch.zeros(*index_shape, *post_index_shape, dtype=torch.float64)
        assert answer.shape == evaluated.shape
        for i in range(index_shape[0]):
            for j in range(index_shape[1]):
                for k in range(index_shape[2]):
                    answer[i, j, k] = x_raw[
                        make_index_at(index_raw[i, j, k], use_index_dim)
                    ]  # of course this doesn't generalise to e.g hypothesis tests : )

        torch.testing.assert_close(answer, evaluated)


@pytest.mark.parametrize("batch_x", [False, True])
def test_gen_index_at_no_index_batch(batch_x: bool):
    torch.manual_seed(1324)
    x_shape = [5, 7]
    index_shape = [1, 4, 2]
    for index_dim in [0, 1, -2, -1]:
        use_index_dim = normalize_index(index_dim, len(x_shape))
        pre_index_shape = x_shape[:use_index_dim]
        post_index_shape = x_shape[use_index_dim + 1 :]
        x_raw = torch.randn(*x_shape, dtype=torch.float64)
        index_raw = torch.randint(low=0, high=x_shape[use_index_dim], size=index_shape)

        x = Array(x_raw, name="x")
        index = Array(index_raw, name="index")

        if batch_x and index_dim >= 0:
            continue

        indexed = GeneralFunction.gen_index(x=x, index=index, index_dim=index_dim, batch_x=batch_x, batch_index=False)
        evaluated = indexed.evaluate()
        assert evaluated.shape == indexed.shape

        answer = torch.zeros(*pre_index_shape, *index_shape, *post_index_shape, dtype=torch.float64)
        assert answer.shape == evaluated.shape
        for i in range(index_shape[0]):
            for j in range(index_shape[1]):
                for k in range(index_shape[2]):
                    answer[(*[slice(None) for _ in pre_index_shape], i, j, k)] = x_raw[
                        make_index_at(index_raw[i, j, k], use_index_dim)
                    ]  # of course this doesn't generalise to e.g hypothesis tests : )

        torch.testing.assert_close(answer, evaluated)


def test_explicit_index():
    torch.manual_seed(1324)
    x_shape = [1, 4, 2, 5, 7]
    index_shape = [1, 4, 2]
    for x_non_batch_dims in range(2, len(x_shape)):
        for index_dim in range(-x_non_batch_dims, x_non_batch_dims):
            x_non_batch_shape = x_shape[-x_non_batch_dims:]
            num_batch_dims = len(x_shape) - x_non_batch_dims
            batch_shape = x_shape[:num_batch_dims]
            index_non_batch_shape = index_shape[num_batch_dims:]
            use_index_dim = normalize_index(index_dim, x_non_batch_dims)
            suffix_post_index_shape = x_non_batch_shape[:use_index_dim] + x_non_batch_shape[use_index_dim + 1 :]
            x_raw = torch.randn(*x_shape, dtype=torch.float64)
            index_raw = torch.randint(low=0, high=x_non_batch_shape[use_index_dim], size=index_shape)

            x = Array(x_raw, name="x")
            index = Array(index_raw.long(), name="index")

            indexed = GeneralFunction.explicit_index(
                x=x, index=index, index_dim=index_dim, x_non_batch_dims=x_non_batch_dims
            )
            evaluated = indexed.evaluate()
            assert evaluated.shape == indexed.shape

            answer = torch.zeros(*batch_shape, *index_non_batch_shape, *suffix_post_index_shape, dtype=torch.float64)
            assert answer.shape == evaluated.shape
            for i in range(index_shape[0]):
                for j in range(index_shape[1]):
                    for k in range(index_shape[2]):
                        answer[i, j, k] = x_raw[
                            (i, j, k)[:num_batch_dims] + make_index_at(index_raw[i, j, k], use_index_dim)
                        ]  # of course this doesn't generalise to e.g hypothesis tests : )

            torch.testing.assert_close(answer, evaluated)



from __future__ import annotations

from copy import copy
from typing import Iterable, Optional, Set, Tuple, Union

import attrs
from attrs import define, frozen

import rust_circuit as rc
from rust_circuit.py_utils import assert_never


@frozen
class ScopeGlobalTraversal:
    traversal: rc.IterativeMatcher


@frozen
class ScopeMatcher:
    matcher: rc.IterativeMatcher


ScopeUpdate = Union[ScopeMatcher, ScopeGlobalTraversal]
NamedUpdates = Tuple[Tuple[ScopeUpdate, Optional[str]], ...]


@define
class ScopeManager:
    """
    Wrapper around a `circuit` for the purposes of doing rewrites

    Note that self.unique() is the circuit focused on, but self.circuit may be
    a parent in a larger computational graph
    """

    circuit: rc.Circuit
    # TODO: add nice stuff for working with names if that seems nice
    named_updates: NamedUpdates = ()
    assert_matched_exists: bool = True
    assert_matched_unique: bool = False

    def __attrs_post_init__(self):
        self.check()

    def check(self):
        matched = self.circuit.get(self.matcher())
        if self.assert_matched_exists:
            assert len(matched) > 0
        if self.assert_matched_unique:
            assert len(matched) == 1

    def raw_matcher(self, *extra: rc.IterativeMatcherIn):
        return rc.IterativeMatcher.new_chain(*self.matchers, *extra)

    def traversal(self, *extra_traversals: rc.IterativeMatcherIn):
        return rc.IterativeMatcher.all(*self.traversals, *extra_traversals)

    def matcher(self, *extra: rc.IterativeMatcherIn, extra_traversals: Iterable[rc.IterativeMatcherIn] = []):
        return self.raw_matcher(*extra) & self.traversal(*extra_traversals)

    @property
    def c(self):
        return self.circuit

    @property
    def matchers(self):
        yield rc.restrict(rc.IterativeMatcher(True), term_if_matches=True)
        for g_m, _ in self.named_updates:
            if isinstance(g_m, ScopeMatcher):
                yield g_m.matcher

    @property
    def traversals(self):
        for g_m, _ in self.named_updates:
            if isinstance(g_m, ScopeGlobalTraversal):
                yield g_m.traversal

    def sub_get_new_updates(self, sub_update: ScopeUpdate, name: Optional[str] = None) -> NamedUpdates:
        """get new named_updates
        sub_item is either node getter for chain or term_early_at item"""
        return self.named_updates + ((sub_update, name),)

    def sub_update(self, sub_update: ScopeUpdate, name: Optional[str] = None) -> ScopeManager:
        """evolve new instance with new named_updates (see sub_u)"""
        return attrs.evolve(self, named_updates=self.sub_get_new_updates(sub_update, name=name))

    def sub_update_(self, sub_update: ScopeUpdate, name: Optional[str] = None):
        """mutably set to new named_updates (see sub_u)"""
        new_named_updates = self.sub_get_new_updates(sub_update, name=name)
        attrs.evolve(self, named_updates=new_named_updates)  # run check
        self.named_updates = new_named_updates

    def sub_matcher(self, matcher: rc.IterativeMatcherIn, name: Optional[str] = None):
        """wrap sub"""
        return self.sub_update(ScopeMatcher(rc.IterativeMatcher(matcher)), name=name)

    def sub_matcher_(self, matcher: rc.IterativeMatcherIn, name: Optional[str] = None):
        """wrap sub_"""
        return self.sub_update_(ScopeMatcher(rc.IterativeMatcher(matcher)), name=name)

    def sub_traversal(self, traversal: rc.IterativeMatcherIn, name: Optional[str] = None):
        """wrap sub"""
        return self.sub_update(ScopeGlobalTraversal(rc.IterativeMatcher(traversal)), name=name)

    def sub_traversal_(self, traversal: rc.IterativeMatcherIn, name: Optional[str] = None):
        """wrap sub_"""
        return self.sub_update_(ScopeGlobalTraversal(rc.IterativeMatcher(traversal)), name=name)

    def remove_by_name(self, name: str):
        return tuple((i, s) for i, s in self.named_updates if s != name)

    def pop_sub_by_name_(self, name: str):
        self.named_updates = self.remove_by_name(name)

    def pop_sub_by_name(self, name: str) -> ScopeManager:
        return attrs.evolve(self, named_updates=self.remove_by_name(name))

    def pop_sub_(self):
        self.named_updates = self.named_updates[:-1]

    def pop_sub(self) -> ScopeManager:
        return attrs.evolve(self, named_updates=self.named_updates[:-1])

    def clear_sub_(self):
        self.named_updates = ()

    def clear_sub(self) -> ScopeManager:
        return attrs.evolve(self, named_updates=())

    def clone(self):
        return copy(self)

    def get_bound_updater(
        self,
        updater: Union[rc.Updater, rc.BoundUpdater, rc.TransformIn],
        apply_global: bool = False,
    ):
        if not isinstance(updater, (rc.Updater, rc.BoundUpdater)):
            updater = rc.Updater(updater)
        if not isinstance(updater, rc.BoundUpdater):
            if apply_global:
                return updater.bind(rc.IterativeMatcher.term(match_next=True))
            else:
                return updater.bind(self.matcher())
        if isinstance(updater, rc.BoundUpdater):
            return self.matcher(updater.matcher).updater(updater.updater.transform)
        else:
            assert_never(updater)

    def update(
        self,
        updater: Union[rc.Updater, rc.BoundUpdater, rc.TransformIn],
        apply_global: bool = False,
    ) -> ScopeManager:
        """evolve new instance with updated circuit (see get_bound_updater)"""
        return attrs.evolve(self, circuit=self.get_bound_updater(updater, apply_global=apply_global)(self.circuit))

    def update_(
        self,
        updater: Union[rc.Updater, rc.BoundUpdater, rc.TransformIn],
        apply_global: bool = False,
    ) -> None:
        """mutably set to new updated circuit (see get_bound_updater)"""
        self.circuit = self.get_bound_updater(updater, apply_global=apply_global)(self.circuit)

    def print_path(
        self,
        printer: rc.PrintOptionsBase = rc.PrintOptions(bijection=False, colorer=lambda _: 3),
        path_from: rc.IterativeMatcherIn = rc.IterativeMatcher.term(match_next=True),
    ):
        """Prints the path to the currently examined nodes.

        Use the colorer of the print passed as argument to color matched nodes"""
        matched = self.matched_circuits()
        path_printer = (
            printer.evolve(
                traversal=self.traversal(rc.new_traversal(term_early_at=~rc.Matcher.match_any_found(matched))),
                colorer=lambda x: (printer.colorer(x) if x in matched else None),  # type: ignore
                bijection=False,
            )
            if isinstance(printer, rc.PrintOptions)
            else printer.evolve(
                traversal=self.traversal(rc.new_traversal(term_early_at=~rc.Matcher.match_any_found(matched))),
                colorer=lambda x: (printer.colorer(x) if x in matched else None),  # type: ignore
            )
        )

        print()
        path_printer.print(rc.Getter().get_unique(self.circuit, path_from))  # type: ignore

    def print(self, printer: rc.PrintOptionsBase = rc.PrintOptions(bijection=False)):
        # TODO: support HTML printer
        printer = (
            printer.evolve(traversal=self.traversal(printer.traversal), bijection=False)
            if isinstance(printer, rc.PrintOptions)
            else printer.evolve(traversal=self.traversal(printer.traversal))
        )
        for c in self.matched_circuits():
            print()
            printer.print(c)

    def unique(self) -> rc.Circuit:
        return self.matcher().get_unique(self.circuit)

    def matched_circuits(self) -> Set[rc.Circuit]:
        return self.matcher().get(self.circuit)



import pytest
import torch

from rust_circuit import Add, Array, Getter, PrintOptions
from rust_circuit.causal_scrubbing.hypothesis import (
    Correspondence,
    ExactSampler,
    FixedOtherSampler,
    IllegalCorrespondenceError,
    InterpNode,
    UncondSampler,
    UncondTogetherSampler,
    corr_root_matcher,
    to_inputs,
)
from rust_circuit.causal_scrubbing.testing_utils import IntDataset


def test_uncond_sampler():
    dataset = IntDataset.of_shape((100,), lambda x: x)
    sampler = UncondSampler()
    sampled: IntDataset = sampler(dataset, dataset)  # type: ignore
    corr = torch.corrcoef(torch.stack((dataset.inp_tensor, sampled.inp_tensor)))[0, 1]
    assert corr < 0.5

    sampled_2: IntDataset = sampler(dataset, dataset)  # type: ignore
    corr_2 = torch.corrcoef(torch.stack((sampled_2.inp_tensor, sampled.inp_tensor)))[0, 1]
    assert corr_2 < 0.5

    assert sampler.ds_eq_class(dataset) == sampler.ds_eq_class(IntDataset.of_shape((100,), lambda x: x))


def test_exact_sampler():
    dataset = IntDataset.of_shape((100, 2), lambda x: x[1])
    other = IntDataset.of_shape((100, 2), lambda x: x[1])
    sampler = ExactSampler()
    sampled = sampler(dataset, other)
    assert sampled == dataset

    assert sampler.ds_eq_class(dataset) != sampler.ds_eq_class(other)


def test_uncond_together_sampler():
    dataset = IntDataset.of_shape((100,), lambda x: x)

    inode = InterpNode(ExactSampler(), "A")
    b0 = inode.make_descendant(UncondTogetherSampler(), "B0")
    b1 = inode.make_descendant(UncondTogetherSampler(), "B1")
    sampled_inputs = inode.sample(None, dataset, dataset)
    assert len(sampled_inputs.sampler_pools) == 1

    sampled: IntDataset = sampled_inputs.datasets[b0]  # type: ignore
    corr = torch.corrcoef(torch.stack((dataset.inp_tensor, sampled.inp_tensor)))[0, 1]
    assert corr < 0.5

    sampled_2 = sampled_inputs.datasets[b1]
    assert sampled_2 == sampled

    other_dataset = IntDataset.of_shape((100,), lambda x: x)
    assert UncondTogetherSampler().ds_eq_class(dataset) == UncondTogetherSampler().ds_eq_class(other_dataset)


def test_fixed_other_sampler():
    dataset = IntDataset.of_shape((100,), lambda x: x)
    other_dataset = IntDataset.of_shape((100,), lambda x: x)

    inode = InterpNode(ExactSampler(), "A")
    b = inode.make_descendant(FixedOtherSampler(other_dataset), "B")
    sampled_inputs = inode.sample(None, dataset, dataset)

    sampled: IntDataset = sampled_inputs.datasets[b]  # type: ignore
    assert sampled == other_dataset


def test_implicit_inputs():
    dataset = IntDataset.of_shape((100,), lambda x: x)
    xs = Array(
        torch.tensor(
            [
                0.0,
            ]
        ),
        name="xs",
    )
    labels = Array(
        torch.tensor(
            [
                0.0,
            ]
        ),
        name="labels",
    )
    C = Add(xs, labels, name="C")

    to_C = corr_root_matcher

    to_inputs_found = Getter().get(C, to_inputs(to_C, dataset))

    assert len(to_inputs_found) == 2
    assert xs in to_inputs_found
    assert labels in to_inputs_found

    dataset = IntDataset(dataset.arrs, input_names={"xs"})

    to_inputs_found = Getter().get(C, to_inputs(to_C, dataset))

    assert len(to_inputs_found) == 1
    assert xs in to_inputs_found
    assert labels not in to_inputs_found


def test_corr_duplicate_keys():
    C_t = InterpNode(name="C", cond_sampler=UncondSampler())

    to_C = corr_root_matcher

    corr = Correspondence()
    with pytest.raises(ValueError):
        corr.replace(C_t, to_C)
    corr.add(C_t, to_C)
    with pytest.raises(ValueError):
        corr.add(C_t, to_C)
    corr.replace(C_t, to_C)


def test_corr_get_root():
    A = InterpNode(name="C", cond_sampler=UncondSampler())
    B = A.make_descendant(name="B", cond_sampler=UncondSampler())

    to_A = corr_root_matcher
    to_B = corr_root_matcher

    corr = Correspondence()
    corr.add(B, to_B)
    corr.add(A, to_A)
    assert A == corr.get_root()


def test_corr_must_be_connected():
    A = InterpNode(name="A", cond_sampler=UncondSampler())
    B = InterpNode(name="B", cond_sampler=UncondSampler())

    to_A = corr_root_matcher
    to_B = corr_root_matcher

    corr = Correspondence()
    corr.add(A, to_A)
    corr.add(B, to_B)
    with pytest.raises(IllegalCorrespondenceError):
        corr.get_root()


def test_corr_must_be_acyclic():
    A = InterpNode(name="C", cond_sampler=UncondSampler())
    B = A.make_descendant(name="B", cond_sampler=UncondSampler())
    B._children = B.children + (A,)  # very bad, no one should ever do this!

    to_A = corr_root_matcher
    to_B = corr_root_matcher

    corr = Correspondence()
    corr.add(A, to_A)
    corr.add(B, to_B)

    with pytest.raises(IllegalCorrespondenceError):
        corr.get_root()


def test_corr_complete():
    A = InterpNode(name="C", cond_sampler=UncondSampler())
    B = A.make_descendant(name="B", cond_sampler=UncondSampler())

    to_A = corr_root_matcher
    to_B = corr_root_matcher

    corr = Correspondence()
    corr.add(A, to_A)
    with pytest.raises(IllegalCorrespondenceError):
        corr.check_complete()

    corr.add(B, to_B)
    corr.check_complete()

    corr = Correspondence()
    corr.add(B, to_B)
    corr.check_complete()


def test_can_sample_interp_graph():
    dataset = IntDataset.of_shape((100,), lambda x: x)

    C_t = InterpNode(name="C", cond_sampler=UncondSampler())
    C_t.make_descendant(name="x0_A", cond_sampler=UncondSampler())
    C_t.make_descendant(name="x0_B", cond_sampler=UncondSampler())

    C_t.sample(source_ds=dataset, parent_ds=dataset, rng=torch.Generator())


def test_can_sample_corr():
    dataset = IntDataset.of_shape((100,), lambda x: x)

    C_t = InterpNode(name="C", cond_sampler=UncondSampler())
    x0_A = C_t.make_descendant(name="x0_A", cond_sampler=UncondSampler())
    x0_B = C_t.make_descendant(name="x0_B", cond_sampler=UncondSampler())

    to_C = corr_root_matcher

    corr = Correspondence()
    corr.add(C_t, to_C)
    corr.add(x0_A, to_C.chain("A").chain("x0"))
    corr.add(x0_B, to_C.chain("B").chain("x0"))

    corr.sample(source_ds=dataset, ref_ds=dataset, rng=torch.Generator())


def test_print():
    C_t = InterpNode(name="C", cond_sampler=ExactSampler())
    C_t.make_descendant(name="x0_A", cond_sampler=ExactSampler())
    C_t.make_descendant(name="x0_B", cond_sampler=UncondSampler())

    before_sampling = """\x1b[90mC GeneralFunction\x1b[0m # cond_sampler=ExactSampler, other_inputs_sampler=UncondSampler
  \x1b[90mx0_A GeneralFunction\x1b[0m # cond_sampler=ExactSampler
  \x1b[90mx0_B GeneralFunction\x1b[0m # cond_sampler=UncondSampler"""

    assert C_t.print(PrintOptions(), repr=True) == before_sampling

    ds = IntDataset((Array(torch.arange(10), "xs"), Array(torch.arange(10), "labels")))
    generator = torch.Generator()
    generator.manual_seed(11)
    sampled_inputs = C_t.sample(generator, ds, ds)

    C_t.print(PrintOptions(), sampled_inputs=sampled_inputs, color_by_data=False)
    after_sampling = """C GeneralFunction # cond_sampler=ExactSampler(d=IntDatum(xs=0, label=0)), other_inputs_sampler=UncondSampler(d=IntDatum(xs=1, label=1))
  x0_A GeneralFunction # cond_sampler=ExactSampler(d=IntDatum(xs=0, label=0))
  x0_B GeneralFunction # cond_sampler=UncondSampler(d=IntDatum(xs=9, label=9))"""

    assert C_t.print(PrintOptions(), sampled_inputs=sampled_inputs, color_by_data=False, repr=True) == after_sampling

    # check coloring doesn't crash, without checking the result. probably we actually want a deterministic colorer arg for testing.
    C_t.print(sampled_inputs=sampled_inputs)



import pytest
import torch

from rust_circuit import Add, Array
from rust_circuit.index_util import Gather, Scatter

#
# Test Gather
#


def test_gather_1():
    specs = [
        "out[i, j] = activations[i, positions[i], j]",  # explicit batching
        "out[i] = activations[positions, i]",  # implicit batching
    ]

    for spec in specs:
        N, M, K = 3, 4, 5

        activations = torch.randn(N, M, K)
        positions = torch.randint(low=0, high=M, size=(N,))

        out_ref = torch.zeros(N, K)
        for i in range(N):
            for j in range(K):
                out_ref[i, j] = activations[i, positions[i], j]

        gather = Gather.new(
            spec,
            {"activations": Array(activations, name="activations"), "positions": Array(positions, name="positions")},
            name="gather",
        )
        out = gather.evaluate()
        torch.testing.assert_close(out_ref, out)


def test_gather_2():
    spec = "out[i, j, k] = x[y[i, j, k], j, k]"

    N, M, K = 3, 4, 5

    x = torch.randn(N, M, K)
    y = torch.randint(low=0, high=N, size=(N, M, K))

    out_ref = torch.zeros(N, M, K)
    for i in range(N):
        for j in range(M):
            for k in range(K):
                out_ref[i, j, k] = x[y[i, j, k], j, k]

    gather = Gather.new(
        spec,
        {"x": Array(x, name="x"), "y": Array(y, name="y")},
        name="gather",
    )
    out = gather.evaluate()
    torch.testing.assert_close(out_ref, out)


def test_gather_3():
    spec = "out[i, j, k] = x[i, y[i, j, k], k]"

    N, M, K = 3, 4, 5

    x = torch.randn(N, M, K)
    y = torch.randint(low=0, high=M, size=(N, M, K))

    out_ref = torch.zeros(N, M, K)
    for i in range(N):
        for j in range(M):
            for k in range(K):
                out_ref[i, j, k] = x[i, y[i, j, k], k]

    gather = Gather.new(
        spec,
        {"x": Array(x, name="x"), "y": Array(y, name="y")},
        name="gather",
    )
    out = gather.evaluate()
    torch.testing.assert_close(out_ref, out)


def test_gather_4():
    spec = "out[i, j, k] = x[i, j, y[i, j, k]]"

    N, M, K = 3, 4, 5

    x = torch.randn(N, M, K)
    y = torch.randint(low=0, high=K, size=(N, M, K))

    out_ref = torch.zeros(N, M, K)
    for i in range(N):
        for j in range(M):
            for k in range(K):
                out_ref[i, j, k] = x[i, j, y[i, j, k]]

    gather = Gather.new(
        spec,
        {"x": Array(x, name="x"), "y": Array(y, name="y")},
        name="gather",
    )
    out = gather.evaluate()
    torch.testing.assert_close(out_ref, out)


def test_gather_5():
    specs = ["out = x[y]", "out[] = x[y]", "out = x[y[]]", "out[] = x[y[]]"]

    N, M, K = 3, 4, 5

    x = torch.randn(N, M, K)
    y = torch.randint(low=0, high=K, size=(N, M))

    out_ref = torch.zeros(N, M)
    for i in range(N):
        for j in range(M):
            out_ref[i, j] = x[i, j, y[i, j]]

    for spec in specs:
        gather = Gather.new(
            spec,
            {"x": Array(x, name="x"), "y": Array(y, name="y")},
            name="gather",
        )

        out = gather.evaluate()
        torch.testing.assert_close(out_ref, out)


def test_gather_6():
    spec = "out[i] = x[i, i, y[i]]"

    N, M = 3, 4

    x = torch.randn(N, N, M)
    y = torch.randint(low=0, high=M, size=(N,))

    out_ref = torch.zeros(N)
    for i in range(N):
        out_ref[i] = x[i, i, y[i]]

    gather = Gather.new(
        spec,
        {"x": Array(x, name="x"), "y": Array(y, name="y")},
        name="gather",
    )
    out = gather.evaluate()
    torch.testing.assert_close(out_ref, out)


def test_gather_7():
    spec = "out[i] = x[i, y[i, i]]"

    N, M = 3, 4

    x = torch.randn(N, M)
    y = torch.randint(low=0, high=M, size=(N, N))

    out_ref = torch.zeros(N)
    for i in range(N):
        out_ref[i] = x[i, y[i, i]]

    gather = Gather.new(
        spec,
        {"x": Array(x, name="x"), "y": Array(y, name="y")},
        name="gather",
    )
    out = gather.evaluate()
    torch.testing.assert_close(out_ref, out)


def test_gather_8():
    spec = "out[i, j] = x[0, j, y[1, i], 2]"

    N0, N1, N2, N3 = 3, 4, 5, 6
    M0, M1 = 7, 8

    x = torch.randn(N0, N1, N2, N3)
    y = torch.randint(low=0, high=N2, size=(M0, M1))

    out_ref = torch.zeros(M1, N1)
    for i in range(M1):
        for j in range(N1):
            out_ref[i, j] = x[0, j, y[1, i], 2]

    gather = Gather.new(
        spec,
        {"x": Array(x, name="x"), "y": Array(y, name="y")},
        name="gather",
    )
    out = gather.evaluate()
    torch.testing.assert_close(out_ref, out)


def test_gather_9():
    spec0 = "out = logits[logit_indices[0]]"
    spec1 = "out = logits[logit_indices[1]]"

    N, M = 8, 10

    logits = torch.randn(N, M)
    logit_indices = torch.randint(low=0, high=M, size=(N, 2))

    out_ref = torch.zeros(N)
    for i in range(N):
        out_ref[i] = logits[i, logit_indices[i, 0]] - logits[i, logit_indices[i, 1]]

    logits_token0 = Gather.new(
        spec0,
        {"logits": Array(logits, name="logits"), "logit_indices": Array(logit_indices, name="logit_indices")},
        name="logits_token0",
    )
    logits_token1 = Gather.new(
        spec1,
        {"logits": Array(logits, name="logits"), "logit_indices": Array(logit_indices, name="logit_indices")},
        name="logits_token1",
    )
    diff = Add.minus(logits_token0, logits_token1, name="diff")
    out = diff.evaluate()
    torch.testing.assert_close(out_ref, out)


def test_gather_10():
    spec = "out[i, j, k] = foo[k, bar[j, k, i], i, j]"

    I, J, K, L = 3, 4, 5, 6

    foo = torch.randn(K, L, I, J)
    bar = torch.randint(low=0, high=L, size=(J, K, I))

    out_ref = torch.zeros(I, J, K)
    for i in range(I):
        for j in range(J):
            for k in range(K):
                out_ref[i, j, k] = foo[k, bar[j, k, i], i, j]

    gather = Gather.new(
        spec,
        {"foo": Array(foo, name="foo"), "bar": Array(bar, name="bar")},
        name="gather",
    )
    out = gather.evaluate()
    torch.testing.assert_close(out_ref, out)


def test_gather_11():
    spec = "out[i, j] = foo[bar[j]]"

    N, M, K = 3, 4, 5

    foo = torch.randn(K)
    bar = torch.randint(low=0, high=K, size=(M,))

    out_ref = torch.zeros(N, M)
    for i in range(N):
        for j in range(M):
            out_ref[i, j] = foo[bar[j]]

    def make_gather(out_shape):
        return Gather.new(
            spec,
            {"foo": Array(foo, name="foo"), "bar": Array(bar, name="bar")},
            name="gather",
            out_shape=out_shape,
        )

    with pytest.raises(Exception, match="Could not infer"):
        make_gather(None)

    for out_shape in [(N, None), (N, M)]:
        out = make_gather(out_shape).evaluate()
        torch.testing.assert_close(out_ref, out)


#
# Test Scatter
#


def test_scatter_1():
    specs = [
        "activations[i, positions[i], j] <- extracted[i, j]",
        "activations[positions, j] <- extracted[j]",
    ]

    N, M, K = 3, 4, 5

    activations = torch.randn(N, M, K)
    positions = torch.randint(low=0, high=M, size=(N,))
    extracted = torch.randn(N, K)

    for reduce in [None, "add", "multiply"]:
        out_ref = activations.clone()
        for i in range(N):
            for j in range(K):
                if reduce is None:
                    out_ref[i, positions[i], j] = extracted[i, j]
                elif reduce == "add":
                    out_ref[i, positions[i], j] += extracted[i, j]
                elif reduce == "multiply":
                    out_ref[i, positions[i], j] *= extracted[i, j]

        for spec in specs:
            scatter = Scatter.new(
                spec,
                {
                    "activations": Array(activations, name="activations"),
                    "positions": Array(positions, name="positions"),
                    "extracted": Array(extracted, name="extracted"),
                },
                name="scatter",
                reduce=reduce,
            )
            out = scatter.evaluate()
            torch.testing.assert_close(out_ref, out)


def test_scatter_2():
    spec = "dst[pos[i, j, k], j, k] <- src[i, j, k]"

    N, M, K = 3, 4, 5

    dst = torch.randn(N, M, K)
    src = torch.randn(N, M, K)
    pos = torch.randint(low=0, high=N, size=(N, M, K))

    for reduce in [None, "add", "multiply"]:
        out_ref = dst.clone()
        for i in range(N):
            for j in range(M):
                for k in range(K):
                    if reduce is None:
                        out_ref[pos[i, j, k], j, k] = src[i, j, k]
                    elif reduce == "add":
                        out_ref[pos[i, j, k], j, k] += src[i, j, k]
                    elif reduce == "multiply":
                        out_ref[pos[i, j, k], j, k] *= src[i, j, k]

        scatter = Scatter.new(
            spec,
            {"dst": Array(dst, name="dst"), "src": Array(src, name="src"), "pos": Array(pos, name="pos")},
            name="scatter",
            reduce=reduce,
        )
        out = scatter.evaluate()
        torch.testing.assert_close(out_ref, out)


def test_scatter_3():
    spec = "dst[i, pos[i, j, k], k] <- src[i, j, k]"

    N, M, K = 3, 4, 5

    dst = torch.randn(N, M, K)
    src = torch.randn(N, M, K)
    pos = torch.randint(low=0, high=M, size=(N, M, K))

    out_ref = dst.clone()
    for i in range(N):
        for j in range(M):
            for k in range(K):
                out_ref[i, pos[i, j, k], k] = src[i, j, k]

    scatter = Scatter.new(
        spec,
        {"dst": Array(dst, name="dst"), "src": Array(src, name="src"), "pos": Array(pos, name="pos")},
        name="scatter",
    )
    out = scatter.evaluate()
    torch.testing.assert_close(out_ref, out)


def test_scatter_4():
    spec = "dst[i, j, pos[i, j, k]] <- src[i, j, k]"

    N, M, K = 3, 4, 5

    dst = torch.randn(N, M, K)
    src = torch.randn(N, M, K)
    pos = torch.randint(low=0, high=K, size=(N, M, K))

    out_ref = dst.clone()
    for i in range(N):
        for j in range(M):
            for k in range(K):
                out_ref[i, j, pos[i, j, k]] = src[i, j, k]

    scatter = Scatter.new(
        spec,
        {"dst": Array(dst, name="dst"), "src": Array(src, name="src"), "pos": Array(pos, name="pos")},
        name="scatter",
    )
    out = scatter.evaluate()
    torch.testing.assert_close(out_ref, out)


def test_scatter_5():
    specs = ["dst[pos] <- src", "dst[pos[]] <- src", "dst[pos] <- src[]", "dst[pos[]] <- src[]"]

    N, M, K = 3, 4, 5

    dst = torch.randn(N, M, K)
    src = torch.randn(N, M)
    pos = torch.randint(low=0, high=K, size=(N, M))

    out_ref = dst.clone()
    for i in range(N):
        for j in range(M):
            out_ref[i, j, pos[i, j]] = src[i, j]

    for spec in specs:
        scatter = Scatter.new(
            spec,
            {"dst": Array(dst, name="dst"), "src": Array(src, name="src"), "pos": Array(pos, name="pos")},
            name="scatter",
        )
        out = scatter.evaluate()
        torch.testing.assert_close(out_ref, out)


def test_scatter_6():
    spec = "dst[i, i, pos[i]] <- src[i]"

    N, M = 3, 4

    dst = torch.randn(N, N, M)
    src = torch.randn(N)
    pos = torch.randint(low=0, high=M, size=(N,))

    for reduce in [None, "add", "multiply"]:
        out_ref = dst.clone()
        for i in range(N):
            if reduce is None:
                out_ref[i, i, pos[i]] = src[i]
            elif reduce == "add":
                out_ref[i, i, pos[i]] += src[i]
            elif reduce == "multiply":
                out_ref[i, i, pos[i]] *= src[i]

        scatter = Scatter.new(
            spec,
            {"dst": Array(dst, name="dst"), "src": Array(src, name="src"), "pos": Array(pos, name="pos")},
            name="scatter",
            reduce=reduce,
        )
        out = scatter.evaluate()
        torch.testing.assert_close(out_ref, out)


def test_scatter_7():
    spec = "dst[i, pos[i, i]] <- src[i, i]"

    N, M = 3, 4

    dst = torch.randn(N, M)
    src = torch.randn(N, N)
    pos = torch.randint(low=0, high=M, size=(N, N))

    for reduce in [None, "add", "multiply"]:
        out_ref = dst.clone()
        for i in range(N):
            if reduce is None:
                out_ref[i, pos[i, i]] = src[i, i]
            elif reduce == "add":
                out_ref[i, pos[i, i]] += src[i, i]
            elif reduce == "multiply":
                out_ref[i, pos[i, i]] *= src[i, i]

        scatter = Scatter.new(
            spec,
            {"dst": Array(dst, name="dst"), "src": Array(src, name="src"), "pos": Array(pos, name="pos")},
            name="scatter",
            reduce=reduce,
        )
        out = scatter.evaluate()
        torch.testing.assert_close(out_ref, out)


def test_scatter_8():
    spec = "dst[0, j, pos[1, i], 2] <- src[i, 3, j]"

    I, J = 3, 4
    N = 5

    dst = torch.randn(2, J, N, 4)
    src = torch.randn(I, 5, J)
    pos = torch.randint(low=0, high=N, size=(3, I))

    for reduce in [None, "add", "multiply"]:
        out_ref = dst.clone()
        for i in range(I):
            for j in range(J):
                if reduce is None:
                    out_ref[0, j, pos[1, i], 2] = src[i, 3, j]
                elif reduce == "add":
                    out_ref[0, j, pos[1, i], 2] += src[i, 3, j]
                elif reduce == "multiply":
                    out_ref[0, j, pos[1, i], 2] *= src[i, 3, j]

        scatter = Scatter.new(
            spec,
            {"dst": Array(dst, name="dst"), "src": Array(src, name="src"), "pos": Array(pos, name="pos")},
            name="scatter",
            reduce=reduce,
        )
        out = scatter.evaluate()
        torch.testing.assert_close(out_ref, out)


def test_scatter_9():
    spec = "dst[i, j, pos[i, j]] <- src[i, j, k]"

    N, I, J, K = 3, 4, 5, 6

    dst = torch.randn(I, J, N)
    src = torch.randn(I, J, K)
    pos = torch.randint(low=0, high=N, size=(I, J))

    with pytest.raises(ValueError, match="missing"):
        Scatter.new(
            spec,
            {"dst": Array(dst, name="dst"), "src": Array(src, name="src"), "pos": Array(pos, name="pos")},
            name="scatter",
        )


def test_scatter_10():
    spec = "dst[i, j, pos[i, j, k, l]] <- src[i, j, k, l]"

    I, J, K, L = 3, 4, 5, 6
    N = 7

    dst = torch.randn(I, J, N)
    src = torch.randn(I, J, K, L)
    pos = torch.randint(low=0, high=N, size=(I, J, K, L))

    with pytest.raises(ValueError, match="missing"):
        Scatter.new(
            spec,
            {"dst": Array(dst, name="dst"), "src": Array(src, name="src"), "pos": Array(pos, name="pos")},
            name="scatter",
        )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional, Tuple, Literal, Callable

from abc import ABC, abstractmethod
from dataclasses import dataclass


class Expr(ABC):
    @abstractmethod
    def render(self, out: list[str]):
        ...

    def __str__(self):
        out = []
        self.render(out)
        return "".join(out)

    def __repr__(self):
        return f"Expr.parse({str(self)!r})"

    @staticmethod
    def parse(s: str) -> "Expr":
        e, str_idx = parse_expr(StrIdx(s, 0))
        if e is None or not str_idx.done():
            raise ValueError(f"Invalid expression: {s!r}")
        return e


@dataclass(repr=False)
class IdentExpr(Expr):
    name: str
    idx: list[Expr]

    def render(self, out: list[str]):
        out.append(self.name)
        if self.idx:
            out.append("[")
            for i, e in enumerate(self.idx):
                if i > 0:
                    out.append(", ")
                e.render(out)
            out.append("]")


@dataclass(repr=False)
class IntExpr(Expr):
    value: int

    def render(self, out: list[str]):
        out.append(str(self.value))


if TYPE_CHECKING:
    EquationKind = Literal["=", "<-"]


@dataclass(repr=False)
class Equation:
    kind: EquationKind
    lhs: Expr
    rhs: Expr

    def render(self, out: list[str]):
        self.lhs.render(out)
        out.append(f" {self.kind} ")
        self.rhs.render(out)

    def __str__(self):
        out = []
        self.render(out)
        return "".join(out)

    def __repr__(self):
        return f"Equation.parse({str(self)!r})"

    @staticmethod
    def parse(s: str) -> "Equation":
        eq, str_idx = parse_equation(StrIdx(s, 0))
        if eq is None or not str_idx.done():
            raise ValueError(f"Invalid equation: {s!r}")
        return eq


@dataclass
class StrIdx:
    content: str
    idx: int

    def advance(self, n: int = 1) -> "StrIdx":
        return StrIdx(self.content, self.idx + n)

    def peek(self, n: int = 1) -> Optional[str]:
        if self.idx + n > len(self.content):
            return None
        return self.content[self.idx : self.idx + n]

    def next(self) -> Tuple[Optional[str], "StrIdx"]:
        c = self.peek()
        if c is None:
            return None, self
        return c, self.advance()

    def skip_whitespace(self) -> "StrIdx":
        while True:
            if self.peek() in [" ", "\t", "\n"]:
                self = self.advance()
            elif self.peek() == "#":
                while self.peek() not in ["\n", None]:
                    self = self.advance()
            else:
                break
        return self

    def done(self) -> bool:
        self = self.skip_whitespace()
        return self.peek() is None


def parse_ident(s: StrIdx) -> Tuple[Optional[str], StrIdx]:
    orig = s
    s = s.skip_whitespace()
    name: list[str] = []
    while (c := s.peek()) is not None and (c.isalnum() or c == "_"):
        if len(name) == 0 and not (c.isalpha() or c == "_"):
            return None, orig
        name.append(c)
        s = s.advance()
    if len(name) == 0:
        return None, orig
    return "".join(name), s


def is_ident(s: str) -> bool:
    ident = parse_ident(StrIdx(s, 0))[0]
    return ident is not None and ident == s


def parse_int(s: StrIdx) -> Tuple[Optional[int], StrIdx]:
    orig = s
    s = s.skip_whitespace()
    value: list[str] = []
    while (c := s.peek()) is not None and c.isdigit():
        value.append(c)
        s = s.advance()
    if len(value) == 0:
        return None, orig
    return int("".join(value)), s


def parse_expr(s: StrIdx) -> Tuple[Optional[Expr], StrIdx]:
    orig = s

    s = s.skip_whitespace()

    ident, s = parse_ident(s)
    if ident is not None:
        idx: list[Expr] = []
        s = s.skip_whitespace()
        if s.peek() != "[":
            return IdentExpr(ident, idx), s
        s = s.advance()
        while True:
            s = s.skip_whitespace()
            if s.peek() == "]":
                s = s.advance()
                break
            e, s = parse_expr(s)
            if e is None:
                return None, orig
            idx.append(e)
            s = s.skip_whitespace()
            if s.peek() == ",":
                s = s.advance()
            elif s.peek() != "]":
                return None, orig
        return IdentExpr(ident, idx), s

    value, s = parse_int(s)
    if value is not None:
        return IntExpr(value), s

    return None, orig


def parse_equation(s: StrIdx) -> Tuple[Optional[Equation], StrIdx]:
    orig = s
    lhs, s = parse_expr(s)
    if lhs is None:
        return None, orig
    s = s.skip_whitespace()
    kind: EquationKind
    if s.peek() == "=":
        kind = "="
    elif s.peek(2) == "<-":
        kind = "<-"
    else:
        return None, orig
    s = s.advance(len(kind))
    rhs, s = parse_expr(s)
    if rhs is None:
        return None, orig
    return Equation(kind, lhs, rhs), s



import hypothesis
import hypothesis.extra.numpy as st_np
import hypothesis.strategies as st
import pytest
import torch
from hypothesis import given, note
from torch.testing import assert_close

import rust_circuit as rc
from interp.circuit.testing.topdown_circuit import CircuitProbs as CP
from interp.circuit.testing.utils import mark_not_interesting_if
from rust_circuit import (
    Index,
    IterativeMatcherIn,
    Module,
    Parser,
    PushDownIndexError,
    PushDownIndexModuleSomeAxesNotPossibleError,
    default_index_traversal,
    new_traversal,
    push_down_index,
    push_down_index_once,
    restrict,
)

from .test_algebric_rewrites import check_transform_equality
from .util import st_Circuit

po = rc.PrintOptions(tensor_index_literal=True, bijection=False)


def run_test_push_down_index_raw(circuit: Index, options, traversal: IterativeMatcherIn = new_traversal()):
    note(circuit.repr(po))
    new_circ = push_down_index(circuit.cast_index(), traversal=traversal, **options)
    if hypothesis.currently_in_test_context():
        hypothesis.assume(new_circ != circuit)
    note(new_circ.repr(po))
    x = circuit.evaluate()
    assert_close(circuit.evaluate(), new_circ.evaluate().type(x.dtype))


index_probs = CP.kw(
    all=0, Index=(1, dict(allow_out_of_bounds_slice=False, allow_none_slice=False, allow_neg_slice=False))
)

options_st = st.fixed_dictionaries(
    dict(
        allow_partial_pushdown=st.booleans(),
        into_module_spec=st.booleans(),
        elim_identity=st.booleans(),
    )
)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=10,
        probs_per_depth=[index_probs],
        probs_default=CP.kw(all=1, Cumulant=0),  # TODO fix bugs and make this 1
        rust=True,
    ),
    options=options_st,
)
@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.filter_too_much])
@mark_not_interesting_if(PushDownIndexError)
def test_push_down_index(circuit: Index, options):
    run_test_push_down_index_raw(circuit, options)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=10,
        probs_per_depth=[index_probs],
        probs_default=CP.kw(all=1, Cumulant=0),
        rust=True,
    ),
    options=options_st,
    end_depth=st.integers(min_value=1, max_value=5),
)
@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.filter_too_much])
@mark_not_interesting_if(PushDownIndexError)
def test_push_down_index_variable_iters(circuit: Index, options, end_depth: int):
    run_test_push_down_index_raw(circuit, options, restrict(default_index_traversal(), end_depth=end_depth))


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=4,
        probs_per_depth=[index_probs, CP.kw(all=0, Concat=1)],
        probs_default=CP.kw(all=1, Cumulant=0),
        rust=True,
    ),
    options=options_st,
)
@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.filter_too_much])
@mark_not_interesting_if(PushDownIndexError)
def test_push_down_index_concat(circuit: Index, options):
    run_test_push_down_index_raw(circuit, options)


@given(
    circuit=st_Circuit(
        st_np.array_shapes(min_dims=0),
        must_be_explicitly_computable=True,
        max_growth_steps=4,
        probs_per_depth=[index_probs, CP.kw(all=0, Einsum=1)],
        probs_default=CP.kw(all=1, Cumulant=0),
        rust=True,
    ),
    options=options_st,
)
@hypothesis.settings(suppress_health_check=[hypothesis.HealthCheck.filter_too_much])
@mark_not_interesting_if(PushDownIndexError)
def test_push_down_index_einsum(circuit: Index, options):
    run_test_push_down_index_raw(circuit, options)


# @given(
#     circuit=st_Circuit(
#         st_np.array_shapes(min_dims=0),
#         must_be_explicitly_computable=True,
#         max_growth_steps=4,
#         probs_per_depth=[index_probs, CP.kw(all=0, Cumulant=1)],
#     )
# )
# @mark_not_interesting_if(PushDownIndexError)
# def test_push_down_index_cumulant(circuit: Index):
#     run_test_push_down_index_raw(circuit)


def test_push_down_index_module_manual():
    s = """
    'spec' Einsum ij,kj,i->kj
      'arg0' [2, 3] Symbol cf812243-3520-40c0-8dc8-56ee05f0fa94
      'arg1' [?, 3] Symbol c2ff6f83-a8cd-4f3e-b16a-476cce27b4c0
      'arg2' [2] Symbol 8f1b8454-43b8-422a-ade0-ee0c1ff136b1

    'a' Module
      'spec'
      'arg0_nb' [2, 3] Array rand ! 'arg0'
      'arg1_nb' [4, 3] Array rand ! 'arg1'
      'arg2_nb' [2] Array rand ! 'arg2'

    'b' Module
      'spec'
      'arg0_b_few' [4, 5, 2, 3] Array rand ! 'arg0'
      'arg1_nb' ! 'arg1'
      'arg2_nb' ! 'arg2'

    'c' Module
      'spec'
      'arg0_b_sing' [7, 2, 3] Array rand ! 'arg0'
      'arg1_nb' ! 'arg1'
      'arg2_b_sing' [7, 2] Array rand ! 'arg2'

    'd' Module
      'spec'
      'arg0_nb' ! 'arg0'
      'arg1_nb' ! 'arg1'
      'arg2_few' [3, 7, 2] Array rand ! 'arg2'

    'e' Module
      'spec'
      'arg0_few' [2, 3] Array rand ! 'arg0'
      'arg1_few' [3, 7, 4, 3] Array rand ! 'arg1'
      'arg2_b_sing' ! 'arg2'

    'f' Module
      'spec'
      'arg0_s_few' [7, 2, 3] Array rand ! 'arg0'
      'arg1_few' ! 'arg1'
      'arg2_b_sing' ! 'arg2'
    """
    spec, *mods = Parser().parse_circuits(s)
    for m in mods:
        num_batch_dims = m.rank - spec.rank

        idxed = m.index([], name="m_idxed")
        out = push_down_index_once(idxed, allow_partial_pushdown=False)
        assert isinstance(out, Module)
        torch.testing.assert_close(out.evaluate(), idxed.evaluate())

        idxed = m.index([0] * num_batch_dims, name="m_idxed")
        out = push_down_index_once(idxed, allow_partial_pushdown=False)
        assert isinstance(out, Module)
        torch.testing.assert_close(out.evaluate(), idxed.evaluate())

        assert push_down_index_once(idxed, allow_partial_pushdown=False) == out
        assert (
            push_down_index_once(
                m.index([*([0] * num_batch_dims), slice(None), slice(None)], name="m_idxed"),
                allow_partial_pushdown=False,
            )
            == out
        )

        idxed_more = m.index([0] * (num_batch_dims + 1), name="m_idxed")

        with pytest.raises(PushDownIndexModuleSomeAxesNotPossibleError):
            push_down_index_once(idxed_more, allow_partial_pushdown=False)

        push_idxed_more = push_down_index_once(idxed_more, allow_partial_pushdown=True)
        assert push_idxed_more.are_any_found(spec)  # still has spec
        assert isinstance(push_idxed_more, Index)

        torch.testing.assert_close(push_idxed_more.evaluate(), idxed_more.evaluate())



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Literal, Optional, Tuple, Sequence, TypeVar, Callable
    from _typeshed import SupportsRichComparisonT
    from rust_circuit.index_util.internal.resolve import IdxExpr

import dataclasses
from dataclasses import dataclass

import torch

import rust_circuit.index_util.internal.infer_shapes as infer_shapes
import rust_circuit.index_util.internal.resolve as resolve
import rust_circuit.index_util.internal.view as view

#
# Planning
#


@dataclass
class OpAxisPlan:
    add_synthetic_axis: bool
    op_axis: int


def plan_op_axis(
    out_rank: int, op_arr_indices: list[IdxExpr], pos_arr_indices: list[IdxExpr]
) -> Tuple[OpAxisPlan, int, list[IdxExpr]]:
    op_idx_set = {idx.axis for idx in op_arr_indices if isinstance(idx, resolve.OutAxisIdx)}
    pos_idx_set = {idx.axis for idx in pos_arr_indices if isinstance(idx, resolve.OutAxisIdx)}

    candidates = pos_idx_set - op_idx_set
    if len(candidates) > 0:
        op_axis = min(candidates)
        return OpAxisPlan(add_synthetic_axis=False, op_axis=op_axis), out_rank, pos_arr_indices

    new_axis = out_rank

    return (
        OpAxisPlan(add_synthetic_axis=True, op_axis=new_axis),
        out_rank + 1,
        pos_arr_indices + [resolve.OutAxisIdx(new_axis)],
    )


@dataclass
class GatherPlan:
    shapes: infer_shapes.ShapesPlan
    axis: OpAxisPlan
    src_view: view.ViewPlan
    pos_view: view.ViewPlan


def plan_gather(spec: resolve.GatherSpec) -> GatherPlan:
    shapes = infer_shapes.plan_shapes(spec.out_rank, [spec.src_indices, spec.pos_indices])

    axis, inner_out_rank, inner_pos_indices = plan_op_axis(spec.out_rank, spec.src_indices, spec.pos_indices)

    src_view_indices: list[view.ViewIdxExpr] = []
    for idx in spec.src_indices:
        if isinstance(idx, resolve.PosDataIdx):
            src_view_indices.append(resolve.OutAxisIdx(axis.op_axis))
        else:
            src_view_indices.append(idx)

    pos_view_indices: list[view.ViewIdxExpr] = []
    for idx in inner_pos_indices:
        assert not isinstance(idx, resolve.PosDataIdx)
        pos_view_indices.append(idx)

    src_view = view.plan_view(inner_out_rank, src_view_indices)
    pos_view = view.plan_view(inner_out_rank, pos_view_indices)

    return GatherPlan(shapes, axis, src_view, pos_view)


@dataclass
class ScatterPlan:
    shapes: infer_shapes.ShapesPlan
    axis: OpAxisPlan
    dst_view: view.ViewPlan
    src_view: view.ViewPlan
    pos_view: view.ViewPlan
    reduce: Optional[Literal["add", "multiply"]]


def plan_scatter(spec: resolve.ScatterSpec) -> ScatterPlan:
    shapes = infer_shapes.plan_shapes(
        len(spec.axis_names),
        [spec.dst_indices, spec.src_indices, spec.pos_indices],
    )

    axis, inner_out_rank, inner_pos_indices = plan_op_axis(len(spec.axis_names), spec.dst_indices, spec.pos_indices)

    dst_view_indices: list[view.ViewIdxExpr] = []
    for idx in spec.dst_indices:
        if isinstance(idx, resolve.PosDataIdx):
            dst_view_indices.append(resolve.OutAxisIdx(axis.op_axis))
        else:
            dst_view_indices.append(idx)

    src_view_indices: list[view.ViewIdxExpr] = []
    for idx in spec.src_indices:
        assert not isinstance(idx, resolve.PosDataIdx)
        src_view_indices.append(idx)
    if axis.add_synthetic_axis:
        src_view_indices.append(resolve.OutAxisIdx(axis.op_axis))

    pos_view_indices: list[view.ViewIdxExpr] = []
    for idx in inner_pos_indices:
        assert not isinstance(idx, resolve.PosDataIdx)
        pos_view_indices.append(idx)

    dst_view = view.plan_view(inner_out_rank, dst_view_indices)
    src_view = view.plan_view(inner_out_rank, src_view_indices)
    pos_view = view.plan_view(inner_out_rank, pos_view_indices)

    return ScatterPlan(shapes, axis, dst_view, src_view, pos_view, spec.reduce)


#
# Execution
#


def infer_gather_shapes(
    plan: GatherPlan,
    out_shape: Optional[list[Optional[int]]],
    src_shape: list[int],
    pos_shape: list[int],
    src_name: str,
    pos_name: str,
) -> infer_shapes.RuntimeShapes:
    return infer_shapes.execute_shapes_plan(
        plan.shapes,
        [(src_name, src_shape), (pos_name, pos_shape)],
        out_shape,
    )


def execute_gather(
    plan: GatherPlan,
    out_shape: Optional[list[Optional[int]]],
    src: torch.Tensor,
    pos: torch.Tensor,
    src_name: str,
    pos_name: str,
) -> torch.Tensor:
    shapes = infer_gather_shapes(plan, out_shape, list(src.shape), list(pos.shape), src_name, pos_name)

    if plan.axis.add_synthetic_axis:
        inner_axis_sizes = shapes.axis_sizes + [1]
        inner_pos = pos.unsqueeze(-1)
    else:
        inner_axis_sizes = shapes.axis_sizes
        inner_pos = pos

    src_view_shape: list[Optional[int]] = list(inner_axis_sizes)
    src_view_shape[plan.axis.op_axis] = None
    src_view = view.execute_plan(plan.src_view, src_view_shape, src)
    pos_view = view.execute_plan(plan.pos_view, inner_axis_sizes, inner_pos)
    src_view = src_view.expand(shapes.batch_shape + list(src_view.shape[src_view.ndim - len(inner_axis_sizes) :]))
    pos_view = pos_view.expand(shapes.batch_shape + inner_axis_sizes)

    inner_gather = torch.gather(src_view, len(shapes.batch_shape) + plan.axis.op_axis, pos_view)

    if plan.axis.add_synthetic_axis:
        return inner_gather.squeeze(-1)
    else:
        return inner_gather


def infer_scatter_shapes(
    plan: ScatterPlan,
    dst_shape: list[int],
    src_shape: list[int],
    pos_shape: list[int],
    dst_name: str,
    src_name: str,
    pos_name: str,
) -> Tuple[infer_shapes.RuntimeShapes, list[int]]:
    shapes = infer_shapes.execute_shapes_plan(
        plan.shapes,
        [(dst_name, dst_shape), (src_name, src_shape), (pos_name, pos_shape)],
    )

    out_shape = torch.broadcast_shapes(shapes.batch_shape + [1] * plan.shapes.arr_ranks[0], dst_shape)

    return shapes, list(out_shape)


def execute_scatter(
    plan: ScatterPlan,
    dst: torch.Tensor,
    src: torch.Tensor,
    pos: torch.Tensor,
    dst_name: str,
    src_name: str,
    pos_name: str,
) -> torch.Tensor:
    shapes = infer_scatter_shapes(
        plan, list(dst.shape), list(src.shape), list(pos.shape), dst_name, src_name, pos_name
    )[0]

    if plan.axis.add_synthetic_axis:
        inner_axis_sizes = shapes.axis_sizes + [1]
        inner_src = src.unsqueeze(-1)
        inner_pos = pos.unsqueeze(-1)
    else:
        inner_axis_sizes = shapes.axis_sizes
        inner_src = src
        inner_pos = pos

    dst = dst.clone()
    dst_view_shape: list[Optional[int]] = list(inner_axis_sizes)
    dst_view_shape[plan.axis.op_axis] = None
    dst_view = view.execute_plan(plan.dst_view, dst_view_shape, dst)
    src_view = view.execute_plan(plan.src_view, inner_axis_sizes, inner_src)
    pos_view = view.execute_plan(plan.pos_view, inner_axis_sizes, inner_pos)
    dst_view = dst_view.expand(shapes.batch_shape + list(dst_view.shape[dst_view.ndim - len(inner_axis_sizes) :]))
    src_view = src_view.expand(shapes.batch_shape + inner_axis_sizes)
    pos_view = pos_view.expand(shapes.batch_shape + inner_axis_sizes)

    dst_view.scatter_(
        len(shapes.batch_shape) + plan.axis.op_axis,
        pos_view,
        src_view,
        **({"reduce": plan.reduce} if plan.reduce is not None else {}),
    )

    return dst



import rust_circuit as rc
import rust_circuit.optional as op
from rust_circuit import (
    Add,
    Array,
    Circuit,
    Concat,
    Cumulant,
    Einsum,
    Index,
    Module,
    Parser,
    Scalar,
    StoredCumulantVar,
    make_broadcast,
    sigmoid,
)
from rust_circuit.module_library import get_pointwise_mlp


def check_auto_names(do_print: bool = False):
    s = Scalar(1.0, (5, 5), "square_mat")
    r = Scalar(1.0, (1, 5), "rect_mat")

    circuits: list[Circuit] = [
        sigmoid(s).add(r),
        sigmoid(s.add(r)),
        Einsum.scalar_mul(Einsum.scalar_mul(r, 2, scalar_name="2"), 3, scalar_name="3"),
        Add(s, r),
        Add(s, op.unwrap(make_broadcast(r, s.shape))),
        op.unwrap(make_broadcast(Index(Add(s, r), [0]), (1, 5))),
        Concat(s, r, axis=0),
        Einsum.scalar_mul(Concat(s, r, axis=0), 2, scalar_name="2"),
        Concat(s, Einsum.scalar_mul(r, 2, scalar_name="2"), axis=0),
        Einsum.scalar_mul(Add(s, r), 2, scalar_name="2"),
        Add(s, Einsum.scalar_mul(r, 2, scalar_name="2")),
        s.sub(r),
        Cumulant(s, Einsum.scalar_mul(r, 2, scalar_name="2"), s),
        StoredCumulantVar({1: s, 2: Cumulant(s, s), 5: Cumulant(*[s for _ in range(5)])}),
        Module.new_flat(
            get_pointwise_mlp().spec,
            Array.randn(12, name="res_stream").add(Array.randn(12, name="noise")),
            Array.randn(12 * 4, 12, name="mlp2.w.in"),
            Array.randn(12 * 4, name="mlp2.w.bias_in"),
            Array.randn(12, 12 * 4, name="mlp2.w.out"),
        ),
        rc.deep_map(
            Parser().parse_circuits(
                """0 Einsum h,oh->o
  1 GeneralFunction gelu
    2 Einsum i,hi->h
      3 Add
        4 'gelu_mlp.input' [6] Symbol
        5 'noise' [6] Symbol
      6 'gelu_mlp.w.proj_in' [3, 6] Symbol
  7 'gelu_mlp.w.proj_out' [6, 3] Symbol"""
            )[0],
            lambda x: x,
        ),
    ]

    for c in circuits:
        assert c.name
        if do_print:
            print(c.name)


def test_auto_names_are_not_none():
    check_auto_names()


if __name__ == "__main__":
    check_auto_names(do_print=True)



# %%
"""
A pool is a group of sampled datasets. If possible, datasets in the pool will
be (randomly) reused when scrubbing instead of sampling new datasets.

What does "if possible" mean? First of all, it has to be enabled by creating a CondSampler
with a PoolAnnotation (by default it is disabled). Then, it has to make semantic sense.
Recall that sampling always happens in the context of 3 things:
- a CondSampler
- a reference datum; the sampler is responsible for sampling a new, "equivalent" datum
- a source dataset to sample from
Suppose a given CondSampler is called on the same ref and source multiple times. Its sample
from the first call could potentially be reused for some of the following calls.

Specifically, we can save the result of calling the CondSampler some number of times, say 4
(this is the pool size), on this ref and source to create a pool. Then, whenever
it's called on the same ref and source, we'll randomly pick a sample from the pool.
"""
from rust_circuit.causal_scrubbing.hypothesis import (
    PNP,
    ExactSampler,
    FuncSampler,
    InterpNode,
    MaybePoolAnnotation,
    PoolAnnotation,
    UncondSampler,
    UncondTogetherSampler,
)
from rust_circuit.causal_scrubbing.testing_utils import IntDataset


class FirstSampler(FuncSampler):
    func = lambda x: IntDataset.unwrap(x).xs.value[:, 0]

    def __init__(self, pool_annot: MaybePoolAnnotation = PNP):
        super().__init__(FirstSampler.func, pool_annot)


dataset = IntDataset.of_shape((1000, 2), lambda x: x[0])

# root node's sampler doesn't have pools enabled
inode = InterpNode(ExactSampler(), "A")
# this sampler has a pool of size 4
first_sampler = FirstSampler(pool_annot=PoolAnnotation(4))
# 40 nodes, each with the same sampler and with the same parent node.
for j in range(40):
    inode.make_descendant(first_sampler, f"B{j}")

sampled_inputs = inode.sample(None, dataset, dataset)

# only one sampler with pools
assert len(sampled_inputs.sampler_pools) == 1
first_sampler_pools = sampled_inputs.sampler_pools[(first_sampler, dataset)]
# only one parent ds this sampler was called on
assert len(first_sampler_pools) == 1
first_sampler_pool_for_parent_ds = list(first_sampler_pools.values())[0]
# pool has size 4
assert len(first_sampler_pool_for_parent_ds) == 4
# so overall we have 5 different sampled datasets: one for the root, 4 for the rest
assert len(set(sampled_inputs.datasets.values())) == 5

inode.print(sampled_inputs=sampled_inputs)
# %%
"""
Why use pools?

This tames the explosion of the treeified model, while only failing to scrub high-order
correlations (that probably you didn't care about anyway). So, this is recommended if your
treeified model is too big (e.g for deep networks with non-trivial samplers).
"""
# %%
"""
In fact, there are additional cases where we could use samples from a pool. The first is:
given a different (by id) sampler, which is the same by attrs eq. This is for convenience, so you
don't have to keep track of all your CondSamplers. As long as you use the same PoolAnnotation
whenever you create CondSamplers, any two that are identical will share a pool (keep in mind
that lambdas are not equal, though, for when you are writing FuncSamplers).
"""
shared_pool = PoolAnnotation(1)
inode = InterpNode(ExactSampler(), "A")
b0 = inode.make_descendant(FirstSampler(pool_annot=shared_pool), "B0")
b1 = inode.make_descendant(FirstSampler(pool_annot=shared_pool), "B1")
assert b0.cond_sampler == b1.cond_sampler

sampled_inputs = inode.sample(None, dataset, dataset)

assert len(sampled_inputs.sampler_pools) == 1
assert len(set(sampled_inputs.datasets.values())) == 2  # one from the root, one from the other two nodes
inode.print(sampled_inputs=sampled_inputs)
# %%
# Meanwhile, samplers with different pool annotations are not equal, and don't share a pool:
inode = InterpNode(ExactSampler(), "A")
b0 = inode.make_descendant(FirstSampler(pool_annot=PoolAnnotation(1)), "B0")
b1 = inode.make_descendant(FirstSampler(pool_annot=PoolAnnotation(1)), "B1")
assert b0.cond_sampler != b1.cond_sampler

sampled_inputs = inode.sample(None, dataset, dataset)

assert len(sampled_inputs.sampler_pools) == 2
assert len(set(sampled_inputs.datasets.values())) == 3  # one from each node
inode.print(sampled_inputs=sampled_inputs)
# %%
# And just for completeness, here's what we get with no pools enabled:
first_sampler = FirstSampler()
inode = InterpNode(ExactSampler(), "A")
b0 = inode.make_descendant(first_sampler, "B0")
b1 = inode.make_descendant(first_sampler, "B1")

sampled_inputs = inode.sample(None, dataset, dataset)

assert len(sampled_inputs.sampler_pools) == 0
assert len(set(sampled_inputs.datasets.values())) == 3  # one from each node
inode.print(sampled_inputs=sampled_inputs)
# %%
"""
The other case where we could reuse samples is: if our sampler is called on a different reference,
which is nonetheless "the same, according to this sampler." Suppose you wrote a sampler which only
cared about the first element in a tuple, and the ref was (1, 2) so the sampler sampled (1, 5). If
the sampler then sees (1, 3), (1, 5) is just as likely to be sampled as it was the first time!

Without this, we might find we rarely make use of our pools, as it might be unlikely for a
sampler to see the same dataset in the course of a causal scrubbing run.

By default, datasets are only equivalent to themselves, i.e. if you extend CondSampler without
overriding ds_eq_class this sort of reuse won't happen. ExactSampler keeps this behavior.
Meanwhile, FuncSampler puts all datasets with the same value of its func in one equivalence class;
and UncondSampler puts all datasets into a single equivalence class.
"""
first_sampler = FirstSampler(pool_annot=PoolAnnotation(1))

inode = InterpNode(ExactSampler(), "A")
# B and C will separately sample data with the same first element
inode_b = inode.make_descendant(FirstSampler(), "B")
inode_c = inode.make_descendant(FirstSampler(), "C")
# So Db and Dc will see different parent ds but with the same first element:
inode_b.make_descendant(first_sampler, "Db")
inode_c.make_descendant(first_sampler, "Dc")

sampled_inputs = inode.sample(None, dataset, dataset)

# only one sampler with pools
assert len(sampled_inputs.sampler_pools) == 1
first_sampler_pools = sampled_inputs.sampler_pools[(first_sampler, dataset)]
# one equivalence class of ds this sampler was called on
eq_class = first_sampler.ds_eq_class(sampled_inputs.datasets[inode_b])
assert eq_class == first_sampler.ds_eq_class(sampled_inputs.datasets[inode_c])
assert len(first_sampler_pools) == 1
first_sampler_pool_for_parent_ds = first_sampler_pools[eq_class]
# pool has size 1
assert len(first_sampler_pool_for_parent_ds) == 1
# so overall we have 4 different sampled datasets: 3 from A B and C, 1 from the pool.
assert len(set(sampled_inputs.datasets.values())) == 4

inode.print(sampled_inputs=sampled_inputs)
# %%
# # Meanwhile if the datasets are "not the same", they do not share a pool:
inode = InterpNode(ExactSampler(), "A")
# B and C will sample totally random data:
inode_b = inode.make_descendant(UncondSampler(), "B")
inode_c = inode.make_descendant(UncondSampler(), "C")
# So Db and Dc will see different parent ds:
inode_b.make_descendant(first_sampler, "Db")
inode_c.make_descendant(first_sampler, "Dc")

sampled_inputs = inode.sample(None, dataset, dataset)

# only one sampler with pools
assert len(sampled_inputs.sampler_pools) == 1
first_sampler_pools = sampled_inputs.sampler_pools[(first_sampler, dataset)]
# two parent ds this sampler was called on
eq_class_b = first_sampler.ds_eq_class(sampled_inputs.datasets[inode_b])
eq_class_c = first_sampler.ds_eq_class(sampled_inputs.datasets[inode_c])
assert eq_class_b != eq_class_c
assert len(first_sampler_pools) == 2
first_sampler_pool_for_b_parent_ds = first_sampler_pools[eq_class_b]
first_sampler_pool_for_c_parent_ds = first_sampler_pools[eq_class_c]
# each pool has size 1
assert len(first_sampler_pool_for_b_parent_ds) == 1
assert len(first_sampler_pool_for_c_parent_ds) == 1
# so overall we have 5 different sampled datasets: 1 for each node
assert len(set(sampled_inputs.datasets.values())) == 5

inode.print(sampled_inputs=sampled_inputs)
# %%
"""
To sum up, we have:
- for each CondSampler hash and source ds:
  - for each ds eq class, according to the sampler:
    - as many samples as the pool size in the sampler's PoolAnnotation specified (in practice
      we sample these lazily, so there's no overhead to using pools when e.g. you only sample once).
"""
# %%
"""
Why use pools? pt 2

Additionally, pools can be used semantially; in particular, a pool of size 1 enforces
things are sampled together.

(And, of course, pools can be used for other_inputs_samplers.)
"""
inode = InterpNode(ExactSampler(), "A")
inode_b = inode.make_descendant(FirstSampler(), "B", other_inputs_sampler=UncondTogetherSampler())
inode_c = inode.make_descendant(FirstSampler(), "C", other_inputs_sampler=UncondTogetherSampler())
# This sampler is always the same by default; but you can pass different uuids to it if desired
assert inode_b.other_inputs_sampler == inode_c.other_inputs_sampler
inode_b.make_descendant(FirstSampler(), "Db")
inode_c.make_descendant(FirstSampler(), "Dc")

sampled_inputs = inode.sample(None, dataset, dataset)

# only one sampler with pools
assert len(sampled_inputs.sampler_pools) == 1
uncond_tog_sampler = UncondTogetherSampler()
uncond_sampler_pools = sampled_inputs.sampler_pools[(uncond_tog_sampler, dataset)]
# according to this sampler, all datasets are in the same equivalence class
eq_class_b = uncond_tog_sampler.ds_eq_class(sampled_inputs.datasets[inode_b])
assert eq_class_b == uncond_tog_sampler.ds_eq_class(sampled_inputs.datasets[inode_c])
assert len(uncond_sampler_pools) == 1
uncond_sampler_pool_for_parent_ds = uncond_sampler_pools[eq_class_b]

# pool has size 1
assert len(uncond_sampler_pool_for_parent_ds) == 1
# so overall we have 2 other_inputs_ds sampled: one from the root, one from the pool
assert len(set(sampled_inputs.other_inputs_datasets.values())) == 2

inode.print(sampled_inputs=sampled_inputs)
# %%
